/************************************************************************/
/* user.c: This file contains routines for errors.                      */
/*                                                                      */
/* Modification Log:                                                    */
/*                                                                      */
/* 91/05/21     Sze-Ying Wuu: Additions/changes made for xdi, v 1.0     */
/* 91/01/11     /f/osi/others/quipu/dish/RCS/user.c,v 7.2               */
/*              Isode Interim (6.8)                               	*/
/************************************************************************/

/* $Header: error.c,v 1.1 91/06/24 14:11:34 sywuu Exp $ */

/*
 *    Original source: ISODE Interim Release 6.8
 *    Changes/additions by: Sze-Ying Wuu, Bellcore, <sywuu@thumper.bellcore.com>
 *
 *               Changes and Additions Copyright (c) 1991 Bellcore
 *                            All Rights Reserved
 *
 *    Permission is granted to copy, modify and redistribute this program
 *    in source or binary form provided that it is not sold for
 *    profit, and all copyright notices, permission and nonwarranty notices 
 *    remain intact.
 *
 *    Bellcore makes no warranty and accepts no liability for this program.
 * 
 *
 *                             ISODE NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */

#include "quipu/util.h"
#include "quipu/read.h"
#include "quipu/sequence.h"

#include "isoaddrs.h"
#include "defs.h"

static struct dua_sequence * current_sequence = NULL_DS;
static struct dua_sequence * top_sequence = NULL_DS;

extern	PS	opt;


/* processing DSE error returned from a search DAP operation */
srch_error (ps,error)
PS ps;
struct DSError * error;
{
struct access_point * ap;
extern char neverefer;
extern int chase_flag;

	if (error->dse_type == DSE_ABANDONED) {
		ps_print (ps,"(DAP call interrupted - abandon successful)\n");
		return (0);
	}

	if (error->dse_type == DSE_ABANDON_FAILED) {
		ps_print (ps,"(DAP call interrupted - abandon unsuccessful)\n");
		return (0);
	}

	if (error->dse_type == DSE_INTRERROR) {
		ps_print (ps,"(DAP call interrupted)\n");
		return (0);
	}

	if ((error->dse_type != DSE_REFERRAL) 
		|| ((chase_flag == 0) && neverefer) 
		|| (chase_flag == 1)) {
		ds_error (ps,error);
		return (0);
	}


	if (error->ERR_REFERRAL.DSE_ref_candidates == NULLCONTINUATIONREF) {
		ps_print (ps,"*** Referral error (but no reference !!!) ***\n");
		return (0);
	}

	for (ap = error->ERR_REFERRAL.DSE_ref_candidates->cr_accesspoints; 
				ap != NULLACCESSPOINT; ap=ap->ap_next) {

		if (chase_flag != 2) {
			ps_print (ps,"Referral to ");
			dn_print (ps,ap->ap_name,EDBOUT);
		}

		if (referral_bind (ap->ap_address) != 0) 
			return (1);

		if (chase_flag == 2)
			break; /* only try first - otherwise possible looping */
	}
	return (0);
}

/* convert DSE error to an xdi error type */
dsEnqError geterr(dserror)
int dserror;
{
    switch (dserror) {
    case DSE_LOCALERROR:
      return(duaerror);
      break;
    case DSE_REMOTEERROR:
      return(localdsaerror);
      break;
    case DSE_ATTRIBUTEERROR:
      return(attributerror);
      break;
    case DSE_REFERRAL:
    case DSE_DSAREFERRAL:
      return(remotedsaerror);
      break;
    case DSE_SECURITYERROR:
      return(security);
      break;
    case DSE_NAMEERROR:
      return(namerror);
      break;
    case DSE_SERVICEERROR:
      return(serviceerror);
      break;
    default:
      return(localdsaerror);
    }
}
