/************************************************************************/
/* filt.c: This file contains routines for formulating filters and	*/
/*         entry attribute selections for Search DAP operations.	*/
/*									*/
/* Modification Log:							*/
/*									*/
/* 91/05/21	Sze-Ying Wuu: Additions/changes made for xdi, Rel 1.0	*/
/* 91/02/22	/f/osi/others/quipu/uips/pod/RCS/filt.c,v 7.2 		*/
/*              Isode Interim (6.8)					*/ 
/************************************************************************/

/* $Header: filt.c,v 1.1 91/06/24 14:30:53 sywuu Exp $ */

/*
 *    Original source: ISODE Interim Release 6.8
 *    Changes/additions by: Sze-Ying Wuu, Bellcore, <sywuu@thumper.bellcore.com>
 *
 *               Changes and Additions Copyright (c) 1991 Bellcore
 *                            All Rights Reserved
 *
 *    Permission is granted to copy, modify and redistribute this program
 *    in source or binary form provided that it is not sold for
 *    profit, and all copyright notices, permission and nonwarranty notices 
 *    remain intact.
 *
 *    Bellcore makes no warranty and accepts no liability for this program.
 * 
 *
 *                             ISODE NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */

#include <string.h>
#include <malloc.h>
#include <ctype.h>

#include "quipu/util.h"
#include "quipu/common.h"
#include "quipu/entry.h"

#include "filt.h"
#include "y.tab.h"
#include "defs.h"

/* external variables */
extern unsigned int curr_filt;
extern unsigned int filt_num;
extern unsigned int typeindx;
extern char *entrytype[];
extern char *default_arr[];
extern PS opt;

/* global variables */
/* These arrays are indexed by entry type and filter/attribute type */
filt_struct *filt_arr[MAXTYPES][MAXTYPES];
attr_type *filttype[MAXTYPES][MAXTYPES];

Attr_Sequence attr_arr[MAXTYPES][MAXTYPES];
attr_type *attrtype[MAXTYPES][MAXTYPES];

/************************************************************************/
/* attr_type	*make_attr_type(attr_name, type_name)			*/
/*		store attribute name and object type name in the  	*/
/* 		structure attr_type.					*/
/************************************************************************/
attr_type *make_attr_type(attr_name, type_name)
char *attr_name, *type_name;
{
  register attr_type *type = (attr_type *) malloc(sizeof(attr_type));

  type->attr_name = (char *) malloc((unsigned int) (strlen(attr_name) + 1));
  type->type_name = (char *) malloc((unsigned int) (strlen(type_name) + 1));

  (void) strcpy(type->attr_name, attr_name);
  (void) strcpy(type->type_name, type_name);
 
  return(type);
}


/************************************************************************/
/* int	make_filt_type(ftype, filt)					*/
/*	store ftype (filter type) in the array filttype and filt 	*/
/*	(filter) in the array filt_arr.					*/	
/************************************************************************/
make_filt_type(ftype, filt)
attr_type *ftype;
filt_struct * filt;
{
  static int curr_attr;

  if (entrytype[curr_filt] == NULLCP) {
     entrytype[curr_filt] = 
          (char *) malloc((unsigned int) (strlen(ftype->type_name) + 1));
     (void) strcpy(entrytype[curr_filt], ftype->type_name);
     curr_attr = 0;
  }

  filttype[curr_filt][curr_attr] = ftype;
  filt_arr[curr_filt][curr_attr++] = filt;

  filttype[curr_filt][curr_attr] = (attr_type *) 0;
  filt_arr[curr_filt][curr_attr] = (filt_struct *) 0;
}

/************************************************************************/
/* int	make_as_type(atype, as)						*/
/*	store atype (attribute sequence type) in the array attrtype 	*/
/*	and as (attribute sequence) in the array attr_arr. 		*/
/************************************************************************/
make_as_type(atype, as)
attr_type *atype;
Attr_Sequence as;
{
  static int curr_attr;
  static int prev_filt = -1;

  if (curr_filt != prev_filt) {
     prev_filt = curr_filt;
     curr_attr = 0;
  }
  attrtype[curr_filt][curr_attr] = atype;
  attr_arr[curr_filt][curr_attr++] = as;

  attrtype[curr_filt][curr_attr] = (attr_type *) 0;
  attr_arr[curr_filt][curr_attr] = NULLATTR;
}

/************************************************************************/
/* filt_struct	*make_as_type(oid, match, value)			*/
/*		formulate filter item as filt_struct for 		*/
/*		(oid match value)					*/
/************************************************************************/
filt_struct *make_item_filter(oid, match, value)
char *oid;
int match;
char *value;
{
  register filt_struct * filt = (filt_struct *) malloc(sizeof(filt_struct));
  
  filt->flt_type = ITEM;
  filt->next = 0;
  
  filt->fu_cont.item.fi_type = match;
  filt->fu_cont.item.stroid = 
    (char *) malloc((unsigned int) (strlen(oid) + 1));
  (void) strcpy(filt->fu_cont.item.stroid, oid);
  
  if (*value == '*') filt->fu_cont.item.name = (char *) 0;
  else {
    filt->fu_cont.item.name = 
      (char *) malloc((unsigned int) (strlen(value) + 1));
    (void) strcpy(filt->fu_cont.item.name, value);
  }
  return filt;
}

/************************************************************************/
/* filt_struct	*link_filters(filt1, filt2)				*/
/*		links filt2 to filt1 and returns filt1.			*/
/************************************************************************/
filt_struct *link_filters(filt1, filt2)
filt_struct *filt1, *filt2;
{
  filt1->next = filt2;
  return filt1;
}

/************************************************************************/
/* filt_struct	*make_parent_filter(filt_type, filt1, filt2, filt3)	*/
/*		creates parent filt_struct for AND/OR/NOT filters .	*/
/************************************************************************/
filt_struct *make_parent_filter(filt_type, filt1, filt2, filt3)
int filt_type;
filt_struct *filt1, *filt2, *filt3;
{
  filt_struct * parent = (filt_struct *) malloc(sizeof(filt_struct));

  switch (filt_type) {

  case NOT:
    parent->flt_type = NOT;
    parent->fu_cont.sub_filt = filt1;
    parent->next = 0;
    break;

  case AND:
    parent->flt_type = AND;
    parent->fu_cont.sub_filt = filt1;
    filt1->next = filt2;
    filt2->next = filt3;
    parent->next = 0;
    break;

  default:
    parent->flt_type = OR;
    parent->fu_cont.sub_filt = filt1;
    filt1->next = filt2;
    filt2->next = filt3;
    parent->next = 0;
    break;
  }

  return parent;
}

/************************************************************************/
/* int	free_filt(filt)							*/
/*	recursively frees filter components 				*/
/************************************************************************/
free_filt(filt)
filt_struct *filt;
{
  if (filt) {
    free_filt(filt->next);

    if (filt->flt_type = ITEM) {
      free(filt->fu_cont.item.stroid);
      if (filt->fu_cont.item.name) free(filt->fu_cont.item.name);
    } else
      free_filt(filt->fu_cont.sub_filt);

    free((char *) filt);
  } else
    return;
}

/************************************************************************/
/* Filter	make_attr_filter(value)					*/
/*		creates filter components				*/
/************************************************************************/
Filter make_attr_filter(value)
char *value;
{
  int match_type;
  char attr_name[STRINGLEN], 
       attr_val[STRINGLEN], sub_val_initial[STRINGLEN], 
       sub_val_final[STRINGLEN], sub_val_any[STRINGLEN];
  register char *end, *start, *next;
  char save;
  Filter sfilt = filter_alloc();

  sfilt->flt_type = FILTER_ITEM;
  sfilt->flt_next = NULLFILTER;
  sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_EQUALITY;
  sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_value = NULLAttrV;
  sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type = NULLAttrT;

  if (end = index(value, '~')) match_type = APPROX;
  else if (end = index(value, '*')) match_type = SUBSTRING;
  else match_type = EQUAL;

  start = value;
  while (isspace(*start) && *start != '\0') start++;
  end = start;
  while (!isspace(*end) && *end != '=' && *end != '~' 
	 && *end != '*' && *end != '\0') end++;
  save = *end;
  *end = '\0';
  (void) strcpy(attr_name, start);
  *end = save;

  if (attr_name [0] == '\0') {
    (void) fprintf (stderr,
	     "Error: Cannot search, invalid syntax on filter '%s'.\n",
	     value);
    filter_free(sfilt);
    return NULLFILTER;
  }

  start = end + 1;
  if (match_type != SUBSTRING) {
    while (!isalnum(*start) && *start != '\0') start++;
    
    end = start;
    
    while (*end != '\0') end++;
    while (!isalnum(*end) && end > start) end--;
    if (*end != '\0') end++;
    
    save = *end;
    *end = '\0';
    (void) strcpy(attr_val, start);
    *end = save;

    if (attr_val[0] == '\0') {
      (void) fprintf (stderr,
	       "Error: Cannot search, invalid syntax on filter '%s'.\n",
	       value);
      filter_free(sfilt);
      return NULLFILTER;
    }
  } else {
    while (!isalnum(*start) && *start != '*' && *start != '\0') start++;
    
    if (*start == '\0') {
      (void) fprintf (stderr,
	       "Error: Cannot search, invalid syntax on filter '%s'.\n",
	       value);
      filter_free(sfilt);
      return NULLFILTER;
    }

    if (*start == '*') {
      sub_val_initial[0] = '*';
      ++start;
      while (isspace(*start) && *start != '\0') start++;
      
      if (*start == '\0' || !isalnum(*start)) {
	(void) fprintf (stderr,
		 "Error: Cannot search, invalid syntax on filter '%s'.\n",
		 value);
	filter_free(sfilt);
	return NULLFILTER;
      }
      
      end = start;
      while (isalnum(*end) && *end != '\0') end++;
      
      if (*end == '\0') {
	(void) strcpy (sub_val_final, start);
	sub_val_any[0] = '*';
      } else {
	next = end;
	
	while (*next != '*' && *next != '\0') next++;
	
	if (*next == '*') {
	  sub_val_final[0] = '*';
	  
	  save = *end;
	  *end = '\0';
	  (void) strcpy(sub_val_any, start);
	  *end = save;
	} else {
	  sub_val_any[0] = '*';
	  
	  save = *end;
	  *end = '\0';
	  (void) strcpy(sub_val_final, start);
	  *end = save;
	}
      }
    } else if (isalnum(*start)) {
      end = start;
      while (!isspace(*end) && *end != '\0') end++;
      
      if (*end == '\0') {
	(void) fprintf (stderr,
		 "Error: Cannot search, invalid syntax on filter '%s'.\n",
		 value);
	filter_free(sfilt);
	return NULLFILTER;
      }
      
      save = *end;
      *end = '\0';
      
      if (index ((char *) (end + 1), '*') == NULLCP) {
	(void) fprintf (stderr,
		 "Error: Cannot search, invalid syntax on filter '%s'.\n",
		 value);
	filter_free(sfilt);
	return NULLFILTER;
      } else {
	(void) strcpy(sub_val_initial, start);
	sub_val_any[0] = sub_val_final[0] = '*';
      }
    }
  }

  switch (match_type) {
  case APPROX:
  case EQUAL:
  case GREATEREQUAL:
  case LESSEQUAL:
    sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type = AttrT_new(attr_name);

      switch (match_type) {
      case APPROX:
        sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_APPROX;
        break;
      case EQUAL:
        sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_EQUALITY;
        break;
      case GREATEREQUAL:
        sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_GREATEROREQUAL;
        break;
      case LESSEQUAL:
        sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_LESSOREQUAL;
        break;
      }

    if (!sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type ||
	sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type->oa_syntax
	== 0) {
      (void) fprintf (stderr,
	       "Error: Cannot search, invalid attribute type '%s'.\n",
	       attr_name);
      sfilt->flt_next = NULLFILTER;
      filter_free(sfilt);
      return NULLFILTER;
    }

    if ((sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_value =
	 str2AttrV(attr_val, 
		   sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type
		   ->oa_syntax)) == NULL) {
      (void) fprintf (stderr,
	 "Error: Cannot search, invalid value '%s' for attribute type '%s'.\n",
	   attr_val, attr_name);
      filter_free(sfilt);
      return NULLFILTER;
    }
    return sfilt;

  case SUBSTRING:
    sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_SUBSTRINGS;

    sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_initial = NULLAV;
    sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_final = NULLAV;
    sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_any = NULLAV;

    sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_type =
      AttrT_new(attr_name);

    if (!sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type ||
	sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type->oa_syntax
	== 0) {
      (void) fprintf (stderr,
	       "Error: Cannot search, invalid attribute type '%s'.\n", 
	       attr_name);
      filter_free(sfilt);
      return NULLFILTER;
    }
    
    if (sub_val_initial[0] != '*')
      if ((sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_initial =
	   avs_comp_new(str2AttrV(sub_val_initial, 
				  sfilt->flt_un.flt_un_item.fi_un.
				  fi_un_substrings.fi_sub_type
				  ->oa_syntax)))
	  == NULLAV) {
	(void) fprintf (stderr,
	 "Error: Cannot search, invalid value '%s' for attribute type '%s'.\n",
		 sub_val_initial, attr_name);
	filter_free(sfilt);
	return NULLFILTER;
      }

    if (sub_val_any[0] != '*')
      if ((sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_any =
	   avs_comp_new(str2AttrV(sub_val_any, 
				  sfilt->flt_un.flt_un_item.fi_un.
				  fi_un_substrings.fi_sub_type->
				  oa_syntax)))
	  == NULLAV) {
	(void) fprintf (stderr,
	 "Error: Cannot search, invalid value '%s' for attribute type '%s'.\n",
		 sub_val_any, attr_name);
	filter_free(sfilt);
	return NULLFILTER;
      }

    if (sub_val_final[0] != '*')
      if ((sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_final =
	   avs_comp_new(str2AttrV(sub_val_final, 
				  sfilt->flt_un.flt_un_item.fi_un.
				  fi_un_substrings.fi_sub_type->
				  oa_syntax)))
	  == NULLAV) {
	(void) fprintf (stderr,
	 "Error: Cannot search, invalid value '%s' for attribute type '%s'.\n",
		 sub_val_final, attr_name);
	filter_free(sfilt);
	return NULLFILTER;
      }
    return sfilt;
  default:
    return NULLFILTER;
  }
}


/************************************************************************/
/* Filter	make_filter_aux(filt, value)				*/
/*		creates real filter for filter components and 		*/
/*		links all filter components 				*/
/************************************************************************/
Filter make_filter_aux(filt, value)
filt_struct *filt;
char *value;
{
  int type;
  char svalue[STRINGLEN], temp[STRINGLEN];
  char *subst;
  Filter rfilt, sfilt;
  
  if (!filt) return NULLFILTER;

  if (index(value, '=')) return make_attr_filter(value);

  sfilt = filter_alloc();

  switch(filt->flt_type) {
  case ITEM:
    sfilt->flt_type = FILTER_ITEM;
    sfilt->flt_next = make_filter_aux(filt->next, value);

    if (filt->fu_cont.item.name)
       if (index(filt->fu_cont.item.name, '$')) {
          strcpy(svalue, filt->fu_cont.item.name);
          if (subst = index(svalue, '*')) {
              strcpy(temp, subst+1); 
              strcpy(subst, value);
              strcat(svalue, temp);
          }
       }
       else
          (void)strcpy(svalue, filt->fu_cont.item.name);
    else 
       (void) strcpy(svalue, value);

    type = filt->fu_cont.item.fi_type;

    switch(type) {
    case APPROX:
    case EQUAL:
    case GREATEREQUAL:
    case LESSEQUAL:
      sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type = 
	AttrT_new(filt->fu_cont.item.stroid);

      if (!sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type ||
	  sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type->
	  oa_syntax == 0) {
        rfilt = sfilt->flt_next;
	sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_value = NULLAttrV;
        sfilt->flt_next = NULLFILTER;
        filter_free(sfilt);
        return rfilt;
      }

      if ((sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_value =
	   str2AttrV(svalue,
		     sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type->
		     oa_syntax)) == NULLAttrV) {
	rfilt = sfilt->flt_next;
	sfilt->flt_next = NULLFILTER;
	filter_free(sfilt);
	return rfilt;
      }

      switch (type) {
      case EQUAL:
	sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_EQUALITY;
        break;
      case APPROX:
	sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_APPROX;
        break;
      case GREATEREQUAL:
	sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_GREATEROREQUAL;
        break;
      case LESSEQUAL:
	sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_LESSOREQUAL;
        break;
      }

      break;

    case SUBSTRING:
      sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_SUBSTRINGS;
      sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_type =
	AttrT_new(filt->fu_cont.item.stroid);
      
      if (!sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type ||
	  sfilt->flt_un.flt_un_item.fi_un.fi_un_ava.ava_type->
	  oa_syntax == 0) {
        rfilt = sfilt->flt_next; 
	sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_initial =
	  NULLAV;
	sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_final = 
	  NULLAV;
	sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_any =
	  NULLAV;
	sfilt->flt_next = NULLFILTER;
        filter_free(sfilt);
        return rfilt;
      }

      sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_initial =
	NULLAV;
      sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_final = 
	NULLAV;
      sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_any =
	avs_comp_new(str2AttrV(svalue,
			       sfilt->flt_un.flt_un_item.fi_un.
			       fi_un_substrings.fi_sub_type->
			       oa_syntax));
      if (sfilt->flt_un.flt_un_item.fi_un.fi_un_substrings.fi_sub_any ==
	  NULLAV) {
	rfilt = sfilt->flt_next;
        sfilt->flt_next = NULLFILTER;
        filter_free(sfilt);
        return rfilt;
      }

      break;
      
    default:
      sfilt->flt_un.flt_un_item.fi_type = FILTERITEM_APPROX;
      break;
    }
    return sfilt;
    
  case AND:
    sfilt->flt_type = FILTER_AND;
    sfilt->flt_un.flt_un_filter = make_filter_aux(filt->fu_cont.sub_filt, value);
    sfilt->flt_next = make_filter_aux(filt->next, value);
    return sfilt;
    
  case OR:
    sfilt->flt_type = FILTER_OR;
    sfilt->flt_un.flt_un_filter = make_filter_aux(filt->fu_cont.sub_filt, value);
    sfilt->flt_next = make_filter_aux(filt->next, value);
    return sfilt;

  case NOT:
    sfilt->flt_type = FILTER_NOT;
    sfilt->flt_next = make_filter_aux(filt->next, value);
    sfilt->flt_un.flt_un_filter = make_filter_aux(filt->fu_cont.sub_filt, value);
    return sfilt;

  default:
    return NULLFILTER;
  }
}


/************************************************************************/
/* Filter	make_filter(filt, attr, value)				*/
/*		get filter from filt_arr based on entry type and 	*/
/*		attr type 						*/
/************************************************************************/
Filter make_filter(filt, attr, value)
Filter filt;
char *attr, *value;
{
  int indx = 0;
  char *al, *nl;
  extern char *strlower();
  Filter sfilt;

  al = strlower(attr);
  nl = NULLCP;

  while (filttype[typeindx][indx]) {
        nl = strlower(filttype[typeindx][indx]->attr_name);
        if (strcmp(al, nl) == 0) break;
        indx++;
  }

  if (filttype[typeindx][indx]) {
     if (filt == NULLFILTER) 
        return(make_filter_aux(filt_arr[typeindx][indx], value));
     else {
        sfilt = filter_alloc();
        sfilt->flt_type = FILTER_AND;
        sfilt->FUFILT = filt;
        sfilt->FUFILT->flt_next = make_filter_aux(filt_arr[typeindx][indx], value);
        sfilt->flt_next = NULLFILTER;
        return(sfilt);
     }
   }
   else 
        return(filt);
 
} 

/************************************************************************/
/* Attr_Sequence	make_as_item(stroid)				*/
/*			creates attribute sequence for stroid		*/
/************************************************************************/
Attr_Sequence make_as_item(stroid)
char *stroid;
{
  AttributeType at;

  if ((at = oid2attr(str2oid(stroid))) != NULLAttrT)
     return(as_comp_new(AttrT_cpy(at))); 
  else
     return(NULLATTR);
}

/************************************************************************/
/* Attr_Sequence	merge_as_seq(as1, as2)				*/
/*			merges as1 to as2 and returns as2		*/
/************************************************************************/
Attr_Sequence merge_as_seq(as1, as2)
Attr_Sequence as1, as2;
{
  if (as2) 
     return(as_merge(as2, as1, NULLAV, NULLACL_INFO));
  else
     return(as1);
}

/************************************************************************/
/* Attr_Sequence	make_as(as, attr)				*/
/*			returns attribute sequence from attr_arr 	*/
/*			based on the object type and attr type		*/
/************************************************************************/
Attr_Sequence make_as(as, attr)
Attr_Sequence as;
char *attr;
{
  int indx = 0;
  char *al, *nl;
  Attr_Sequence eptr;

  al = strlower(attr);
  nl = NULLCP;

  while (attrtype[typeindx][indx]) {
       nl = strlower(attrtype[typeindx][indx]->attr_name);
       if (strcmp(al, nl) == 0) break;
       indx++;
  }
  
  if (attr_arr[typeindx][indx])
     for (eptr = attr_arr[typeindx][indx]; eptr != NULLATTR; eptr= eptr->attr_link)
         as = as_merge(as, as_comp_new(eptr->attr_type), NULLAV, NULLACL_INFO);

  return(as);
}

